# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.

import torch
import pytest
from aiter.ops.triton.attention.prefill_attention import (
    context_attention_fwd as prefill_triton,
)


def input_helper(BATCH, SEQLEN, H, HEAD_DIM, dtype, absorb=False):
    if absorb:
        q = torch.randn((BATCH * SEQLEN, H, 512), device="cuda", dtype=dtype)
        k = torch.randn((BATCH * SEQLEN, 1, 512), device="cuda", dtype=dtype)
        v = torch.randn((BATCH * SEQLEN, 1, 512), device="cuda", dtype=dtype)
    else:
        q = torch.randn((BATCH * SEQLEN, H, HEAD_DIM), device="cuda", dtype=dtype)
        k = torch.randn((BATCH * SEQLEN, H, HEAD_DIM), device="cuda", dtype=dtype)
        v = torch.randn((BATCH * SEQLEN, H, HEAD_DIM), device="cuda", dtype=dtype)

    b_seq_len = torch.ones((BATCH,), dtype=torch.int32, device="cuda") * SEQLEN
    b_start_loc = torch.zeros((BATCH,), dtype=torch.int32, device="cuda")

    b_start_loc = torch.cat(
        [
            torch.tensor([0], dtype=torch.int32, device="cuda"),
            b_seq_len.cumsum(dim=0, dtype=torch.int32),
        ]
    )

    return q, k, v, b_seq_len, b_start_loc[:BATCH]


def varlen_input_helper(
    BATCH, SEQLEN, H, HEAD_DIM, dtype, absorb=False, equal_seqlens=False
):
    if not equal_seqlens:
        max_seqlens = SEQLEN // BATCH
        seqlens = torch.randint(
            1, max_seqlens + 1, (BATCH,), dtype=torch.int32, device="cuda"
        )
    else:
        seqlens = torch.full((BATCH,), SEQLEN // BATCH, device="cuda")

    # Calculate cumulative sequence lengths
    cu_seqlens = torch.cat(
        [
            torch.tensor([0], dtype=torch.int32, device="cuda"),
            seqlens.cumsum(dim=0, dtype=torch.int32),
        ]
    )

    # Initialize q, k, v with variable lengths
    total_seq = cu_seqlens[-1].item()

    if absorb:
        q = torch.randn((total_seq, H, 512), device="cuda", dtype=dtype)
        k = torch.randn((total_seq, 1, 512), device="cuda", dtype=dtype)
        v = torch.randn((total_seq, 1, 512), device="cuda", dtype=dtype)
    else:
        q = torch.randn((total_seq, H, HEAD_DIM), device="cuda", dtype=dtype)
        k = torch.randn((total_seq, H, HEAD_DIM), device="cuda", dtype=dtype)
        v = torch.randn((total_seq, H, HEAD_DIM), device="cuda", dtype=dtype)

    return q, k, v, seqlens, cu_seqlens[:BATCH]


@pytest.mark.parametrize(
    "Z, H, SEQLEN, HEAD_DIM",
    [
        (4, 48, 1024, 64),
        (1, 24, 8192, 64),
        (1, 4, 16384, 128),
        (2, 16, 1020, 128),
        (2, 16, 15498, 128),
        (4, 48, 1001, 64),
        (1, 8, 8081, 64),
        (1, 4, 16330, 128),
    ],
)
@pytest.mark.parametrize("causal", [True, False])
@pytest.mark.parametrize("absorb", [False])
@pytest.mark.parametrize("varlen", [True, False])
def test_op_fwd(Z, H, SEQLEN, HEAD_DIM, causal, absorb, varlen, dtype=torch.float16):
    torch.cuda.empty_cache()  # Helps avoid hangs in large tests
    torch.manual_seed(20)
    if varlen:
        q, k, v, b_seq_len, b_start_loc = varlen_input_helper(
            Z, SEQLEN, H, HEAD_DIM, dtype, absorb
        )
    else:
        q, k, v, b_seq_len, b_start_loc = input_helper(
            Z, SEQLEN, H, HEAD_DIM, dtype, absorb
        )

    tri_out = torch.empty_like(q)
    ref_out = torch.empty_like(q)

    # triton implementation
    prefill_triton(q, k, v, tri_out, b_start_loc, b_seq_len, SEQLEN, causal)
    sm_scale = 1.0 / (HEAD_DIM**0.5)

    for i in range(0, Z):
        start_q, start_k = b_start_loc[i], b_start_loc[i]
        seqlen_q, seqlen_k = b_seq_len[i], b_seq_len[i]
        end_q, end_k = start_q + seqlen_q, start_k + seqlen_k
        scores = torch.einsum(
            "qhd,khd->qhk", q[start_q:end_q], k[start_k:end_k]
        ).float()
        if causal:
            # Apply causal mask to scores (set future positions to -inf)
            # Create a causal mask
            causal_mask = torch.triu(
                torch.ones(seqlen_q, seqlen_k, device=scores.device), diagonal=1
            ).bool()
            scores.masked_fill_(causal_mask.unsqueeze(1), float("-inf"))
        scores = scores.float()

        p = torch.softmax(scores * sm_scale, dim=-1).half()
        ref_out[start_q:end_q] = torch.einsum("qhk,khd->qhd", p, v[start_k:end_k])

    # compare
    torch.testing.assert_close(ref_out, tri_out, atol=2e-2, rtol=2e-2)



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.

import torch
import aiter
from torch.profiler import profile, ProfilerActivity
from aiter import dtypes
import argparse

parser = argparse.ArgumentParser(
    formatter_class=argparse.RawTextHelpFormatter,
    description="config input of test",
)
parser.add_argument(
    "-s",
    "--shape",
    nargs="*",
    type=dtypes.str2tuple,
    choices=[
        (512,),
        (1280, 232, 256),
        (256, 256),
        (256, 8192),
        (256,),
        (1280, 32, 256),
        (384, 256),
        (384,),
        (65536,),
        (65536, 256),
        (1, 8, 256),
        (512, 256),
        (1280, 532, 256),
    ],
    default=[
        (512,),
        (1280, 232, 256),
        (256, 256),
        (256, 8192),
        (256,),
        (1280, 32, 256),
        (384, 256),
        (384,),
        (65536,),
        (65536, 256),
        (1, 8, 256),
        (512, 256),
        (1280, 532, 256),
    ],
    help="""Input shapes.  Dimensionality must match dimensionality of stride (-st).
    e.g.: -s 1280,232,256""",
)
parser.add_argument(
    "-st",
    "--stride",
    nargs="*",
    type=dtypes.str2tuple,
    choices=[
        (1,),
        (59392, 256, 1),
        (256, 1),
        (8192, 1),
        (1,),
        (8192, 256, 1),
        (256, 1),
        (1,),
        (1,),
        (256, 1),
        (2048, 256, 1),
        (256, 1),
        (136192, 256, 1),
    ],
    default=[
        (1,),
        (59392, 256, 1),
        (256, 1),
        (8192, 1),
        (1,),
        (8192, 256, 1),
        (256, 1),
        (1,),
        (1,),
        (256, 1),
        (2048, 256, 1),
        (256, 1),
        (136192, 256, 1),
    ],
    help="""Input strides. Dimensionality must match dimensionality of shape (-s).
    e.g.: -st 59392,256,1""",
)
# input shape: torch.Size([4096, 64, 160]) (20480, 1, 128)
# other shape: torch.Size([4096, 64, 160]) (10240, 160, 1)

# input shape: torch.Size([4096, 64, 160]) (47360, 1, 296)
# other shape: torch.Size([4096, 64, 160]) (10240, 160, 1)

# shape0 = (4096, 64, 160)
# shape1 = (4096, 64, 160)
# stride0 = (47360, 1, 296)
# # stride1 = (20480, 1, 128)
# stride1 = (10240, 160, 1)

# shape1 = (4096, 200, 64)
# shape0 = (1, 200, 64)
# stride0 = (12800, 64, 1)
# stride1 = (12800, 64, 1)

# shape1 = (144, 1, 160)
# shape0 = (144, 4096, 160)
# stride1 = (160, 160, 1)
# stride0 = (655360, 160, 1)

args = parser.parse_args()

tensors0 = [
    torch.empty_strided(shape, stride, dtype=dtypes.bf16, device="cuda")
    for shape, stride in zip(args.shape, args.stride)
]
tensors1 = [
    torch.empty_strided(shape, stride, dtype=dtypes.bf16, device="cuda")
    for shape, stride in zip(args.shape, args.stride)
]
for tensor in tensors0:
    tensor.copy_(torch.rand_like(tensor))
    # tensor.fill_(1)
for tensor in tensors1:
    tensor.copy_(torch.rand_like(tensor))
    # tensor.fill_(1)

# tensor0 = torch.empty_strided(shape0, stride0, dtype=dtypes.bf16, device='cuda')
# tensor1 = torch.empty_strided(shape1, stride1, dtype=dtypes.bf16, device='cuda')
# # tensor0 = torch.empty_strided(shape0, stride0, dtype=dtypes.fp32, device='cuda')
# # tensor1 = torch.empty_strided(shape1, stride1, dtype=dtypes.fp32, device='cuda')
# random_data0 = torch.rand(shape0)
# # tensor0.copy_(random_data0)
# tensor0.fill_(0)
# random_data1 = torch.rand(shape1)
# # tensor1.copy_(random_data1)
# tensor1.fill_(2)

for tensor0, tensor1 in zip(tensors0, tensors1):
    print("shape:", tensor0.size())
    with profile(
        activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
        profile_memory=True,
        with_stack=True,
        with_modules=True,
        record_shapes=True,
    ) as prof:
        for j in range(100):
            # cache_flush1 = torch.randn(10000, 10000, requires_grad=True, device="cuda", dtype=dtypes.fp32).to(dtypes.i32)
            # result = torch.add_(tensor0, tensor1)
            torch_add_a_ = tensor0.clone()
            torch_add_b_ = tensor1.clone()
            torch_add_a_.add_(torch_add_b_)
            result = torch_add_a_
            # result_con = result.contiguous()
    print(prof.key_averages().table(sort_by="cuda_time_total", row_limit=10))

    with profile(
        activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
        profile_memory=True,
        with_stack=True,
        with_modules=True,
        record_shapes=True,
    ) as prof:
        for j in range(100):
            # cache_flush1 = torch.randn(10000, 10000, requires_grad=True, device="cuda", dtype=dtypes.fp32).to(dtypes.i32)
            # output = torch.empty_like(tensor1)
            aiter_add_a_ = tensor0.clone()
            aiter_add_b_ = tensor1.clone()
            output = aiter.add_(aiter_add_a_, aiter_add_b_)
    print(prof.key_averages().table(sort_by="cuda_time_total", row_limit=10))

    print(torch.equal(result, output))
    # print("result:", result)
    # print("output:", output)



from collections import namedtuple
import os
import concurrent.futures
from csrc.cpp_itfs.sampling.top_k_renorm_probs import (
    compile as top_k_renorm_probs_compile,
)
from csrc.cpp_itfs.sampling.top_p_sampling_from_probs import (
    compile as top_p_sampling_from_probs_compile,
)
from csrc.cpp_itfs.sampling.top_k_top_p_sampling_from_probs import (
    compile as top_k_top_p_sampling_from_probs_compile,
)

TopKRenormConfig = namedtuple(
    "TopKRenormConfig",
    ["vec_size", "func_name"],
)

TopPSamplingConfig = namedtuple(
    "TopPSamplingConfig",
    ["vec_size", "deterministic", "func_name"],
)

TopKTopPSamplingConfig = namedtuple(
    "TopKTopPSamplingConfig",
    ["vec_size", "deterministic", "func_name"],
)


def process_top_k_renorm_config(config):
    return top_k_renorm_probs_compile(config.vec_size)


def process_top_p_sampling_config(config):
    return top_p_sampling_from_probs_compile(config.vec_size, config.deterministic)


def process_top_k_top_p_sampling_config(config):
    return top_k_top_p_sampling_from_probs_compile(
        config.vec_size, config.deterministic
    )


def main():
    # Generate configs for top_k_renorm_probs
    top_k_renorm_configs = []
    for vec_size in range(1, 5):
        top_k_renorm_configs.append(
            TopKRenormConfig(
                vec_size=vec_size,
                func_name="top_k_renorm_probs",
            )
        )

    # Generate configs for top_p_sampling_from_probs
    top_p_sampling_configs = []
    for vec_size in range(1, 5):
        for deterministic in [False, True]:
            top_p_sampling_configs.append(
                TopPSamplingConfig(
                    vec_size=vec_size,
                    deterministic=deterministic,
                    func_name="top_p_sampling_from_probs",
                )
            )

    # Generate configs for top_k_top_p_sampling_from_probs
    top_k_top_p_sampling_configs = []
    for vec_size in range(1, 5):
        for deterministic in [False, True]:
            top_k_top_p_sampling_configs.append(
                TopKTopPSamplingConfig(
                    vec_size=vec_size,
                    deterministic=deterministic,
                    func_name="top_k_top_p_sampling_from_probs",
                )
            )

    max_jobs = int(os.environ.get("MAX_JOBS", os.cpu_count() or 16))

    # Process all configs in parallel
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_jobs) as executor:
        executor.map(process_top_k_renorm_config, top_k_renorm_configs)
        executor.map(process_top_p_sampling_config, top_p_sampling_configs)
        executor.map(process_top_k_top_p_sampling_config, top_k_top_p_sampling_configs)


if __name__ == "__main__":
    main()
